"""
*******************************************************
*
*  plot_marginals - PLOT MARGINALS
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   September 10, 2018
*  Last updated: September 18, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
import matplotlib.pyplot as plt


#################################################################
#  PLOT MARGINAL DISTRIBUTION
#################################################################
def plot_marginals(metrics):
    """Plots marginal distributions for each parameter of 
    the fitted model.
    
    Keyword arguments:
    metrics -- contain important metrics about fitted model (dictionary)
    """
    # Generate basic plot of marginal distributions
    fig, axes = plt.subplots(2, 2, 
                             subplot_kw=dict(polar=False), 
                             figsize = (7,6))
    # Scale parameter
    axes[0,0].set_xlabel('Scale')
    axes[0,1].set_xlabel('Slope')
    axes[1,0].set_xlabel('Gamma')
    axes[1,1].set_xlabel('Lambda')
    # Loop through and plot marginals that exist
    counter = 0
    idx = np.array([[0,0], [0,1], [1,0], [1,1]])
    for keys in ['scale', 'slope', 'gamma', 'lambda']:
        axes[idx[counter,0],idx[counter,1]].set_ylabel('Probability')
        if metrics['Marginals'][keys] is not np.nan and metrics['Marginals'][keys].size > 1:
            axes[idx[counter,0],idx[counter,1]].plot(metrics['Marginals_X'][keys],
                metrics['Marginals'][keys], 
                lw=3, 
                color='#5998ff')
            axes[idx[counter,0],idx[counter,1]].fill_between(metrics['Marginals_X'][keys],
                metrics['Marginals'][keys], color='#5998ff', alpha = .4)
        elif metrics['Marginals'][keys].size == 1:
            axes[idx[counter,0],idx[counter,1]].text(0.5,0.5, "None",
                horizontalalignment='center', 
                verticalalignment='center', 
                transform=axes[idx[counter,0],idx[counter,1]].transAxes)
        # Update counter
        counter += 1
    plt.tight_layout()
    plt.show()



"""
*******************************************************
*
*  extractPriors - EXTRACT PRIORS PROVIDED BY USER 
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   May 19, 2018
*  Last updated: February 17, 2024
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
import re 


#################################################################
#  EXTRACT LIST OF PRIORS PROVIDED
#################################################################
# Extract priors from options
def extract_priors(options): 
    """Extract priors from user definitions using regular
    expressions, and store under new keys in options.

    Keyword arguments:
    options -- contains all options used to fit model (dictionary)
    """
    # Initialize empty dictionary to store priors 
    options['priors_definitions'] = dict()
    options['priors_params'] = dict()

    if len(options['priors']) < 4:
        raise ValueError('''Please provide list of four arguments for 
                            priors. Less than four arguments provided.''')
    elif len(options['priors']) > 4:
        raise ValueError('''Please provide list of four arguments for 
                            priors. More than four arguments provided.''')
    elif len(options['priors']) == 4: 
        NoneType = type(None)
        # Check whether arguments in list are of type integer or float
        for i in range(0,4):
            if isinstance(options['priors'][i], (str, NoneType)) is False:
                raise ValueError('''User Error: Please provide numerical values or NoneType 
                                 for priors of parameters of the model.''')
        
            if options['priors'][i] is not None:
                if i == 0:
                    # Extract prior for alpha parameter
                    options['priors_definitions']['scale'] = " ".join(re.findall("[a-zA-Z]+", 
                                                                      options['priors'][0]))
                    
                    alpha_prior_values = re.findall(r'[-+]?\d*\.\d+|\d+', options['priors'][0])
                    options['priors_params']['scale'] = [float(x) for x in alpha_prior_values]
                elif i == 1:
                    # Extract prior for beta parameter
                    options['priors_definitions']['slope'] = " ".join(re.findall("[a-zA-Z]+", 
                                                                     options['priors'][1]))
                    
                    beta_prior_values = re.findall(r'[-+]?\d*\.\d+|\d+', options['priors'][1])
                    options['priors_params']['slope'] = [float(x) for x in beta_prior_values]    
                elif i == 2:
                    # Extract prior for gamma parameter
                    options['priors_definitions']['gamma']  = " ".join(re.findall("[a-zA-Z]+", 
                                                                       options['priors'][2]))  
                    gamma_prior_values = re.findall(r'[-+]?\d*\.\d+|\d+', options['priors'][2])
                    options['priors_params']['gamma'] = [float(x) for x in gamma_prior_values]
                elif i == 3:
                    # Extract prior for lambda parameter
                    options['priors_definitions']['lambda']  = " ".join(re.findall("[a-zA-Z]+", 
                                                                        options['priors'][3]))
            
                    lambda_prior_values = re.findall(r'[-+]?\d*\.\d+|\d+', options['priors'][3])
                    options['priors_params']['lambda'] = [float(x) for x in lambda_prior_values]
            
        return options



"""
*******************************************************
*
*  test_checkData - UNIT TEST FOR TRAVIS CI
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   April 28, 2017
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES 
#################################################################
import numpy as np
from .checkData import check_data as _check_data


#################################################################
#  DEFINE FUNCTIONS USED FOR UNIT TESTING
#################################################################

def _batch_false(error):    
    # Test cases for batch equal False 
    batch = False    
    # Generate dataset 
    data = np.array([[0.0010,   45.0000,   90.0000],
                     [0.0015,   50.0000,   90.0000],
                     [0.0020,   44.0000,   90.0000],
                     [0.0025,   44.0000,   90.0000],
                     [0.0030,   52.0000,   90.0000],
                     [0.0035,   53.0000,   90.0000],
                     [0.0040,   62.0000,   90.0000],
                     [0.0045,   64.0000,   90.0000],
                     [0.0050,   76.0000,   90.0000],
                     [0.0060,   79.0000,   90.0000],
                     [0.0070,   88.0000,   90.0000],
                     [0.0080,   90.0000,   90.0000],
                     [0.0100,   90.0000,   90.0000]]);
    # Run through possible error types
    if error == 0:
        data = data
    elif error == 1:
        data = dict()
    elif error == 2:
        data = data[:,1:2]
    elif error == 3:
        data[0,2] = 0 
    # Call function with arguments above
    _check_data(data, batch)
    # Update success flag
    success = 1
    return success 


def _batch_true(error):    
    # Test cases for batch equal False 
    batch = True    
    # Generate dataset 
    data_batch = np.array([[0.0010,   45.0000,   90.0000],
                             [0.0015,   50.0000,   90.0000],
                             [0.0020,   44.0000,   90.0000],
                             [0.0025,   44.0000,   90.0000],
                             [0.0030,   52.0000,   90.0000],
                             [0.0035,   53.0000,   90.0000],
                             [0.0040,   62.0000,   90.0000],
                             [0.0045,   64.0000,   90.0000],
                             [0.0050,   76.0000,   90.0000],
                             [0.0060,   79.0000,   90.0000],
                             [0.0070,   88.0000,   90.0000],
                             [0.0080,   90.0000,   90.0000],
                             [0.0100,   90.0000,   90.0000]]);
    # Run through possible error types
    if error == 0:
        data = dict()
        data['obs1'] = data_batch
    elif error == 1:
        data = list()
    elif error == 2:
        data = dict()
        data['obs1'] = data_batch[:,1:2]
    elif error == 3:
        data = dict()
        data['obs1'] = data_batch
        data['obs1'][0,2] = 0
    # Call function with arguments above
    _check_data(data, batch)
    # Update success flag
    success = 1
    return success 


#################################################################
#  UNIT TESTS
#################################################################

def test_batch_false_noerror():
    success = _batch_false(0)
    # Assert if exception  
    assert success == 1

def test_batch_false_error1():
    raised = False
    try:
        _batch_false(1)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_batch_false_error2():
    raised = False
    try:
        _batch_false(2)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True

def test_batch_false_error3():
    raised = False
    try:
        _batch_false(3)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True


def test_batch_true_noerror():
    success = _batch_true(0)
    # Assert if exception  
    assert success == 1

def test_batch_true_error1():
    raised = False
    try:
        _batch_true(1)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_batch_true_error2():
    raised = False
    try:
        _batch_true(2)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True

def test_batch_true_error3():
    raised = False
    try:
        _batch_true(3)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True



"""
*******************************************************
*
*  applyPriors - APPLY PRIORS TO LIKELIHOOD SURFACE TO FORM POSTERIOR
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   September 3, 2018
*  Last updated: November 19, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
import scipy.stats as st


#################################################################
#  FUNCTION THAT DEFINES DISTRIBUTION BASED ON CLASS
#################################################################
def _define_distribution(xx, user_defined_distro, distro_parameters):
    """Assigns probability density to parameters defined by grid
    and specified distribution.

    Keyword arguments:
    xx -- range of values over which to define distribution (ndarray)
    user_defined_distro -- name of distribution (string)
    distro_parameters -- parameters for distribution (list)
    """
    # Define distrobution based on inputs
    if user_defined_distro == 'Unif':
        distro = st.uniform.pdf(xx,
                                distro_parameters[0],
                                distro_parameters[1])
    elif user_defined_distro == 'Norm':
        distro = st.norm.pdf(xx,
                             distro_parameters[0],
                             distro_parameters[1])
    elif user_defined_distro == 'Log-Norm':
        distro = st.lognorm.pdf(xx,
                                distro_parameters[0],
                                distro_parameters[1])
    elif user_defined_distro == 'Beta':
        distro = st.beta.pdf(xx,
                             distro_parameters[0],
                             distro_parameters[1])
    elif user_defined_distro == 'Gamma':
        distro = st.gamma.pdf(xx,
                              distro_parameters[0],
                              distro_parameters[1])

    # Check whether vector or probabilities returns non-zero sum
    if distro.sum() == 0.0:
        raise ValueError('''Error applying priors: A vector representing
            the prior probabilities along the distribution for one of the
            parameters you have specified sums to zero. This suggests one of
            the prior distributions provided are not within the range of, or
            far off from those regions where the best estimated value
            likely exists.
            \n
            Please adjust the priors provided to capture regions where
            value estimates are more probable, or set priors to
            None to obtain use MLE.''')
    else:
        # Ensure distribution sums to one
        distro /= distro.sum()

    return distro


#################################################################
#  APPLY PRIORS
#################################################################
def apply_priors(data, options, posterior, grid):
    """Applies prior to likelihood surface resulting in
    use of Bayesian inference to produce posterior surface.

    Keyword arguments:
    data -- m x 3 numpy array
    options -- contains all options used to fit model (dictionary)
    posterior -- posterior surface (ndarray)
    grid -- grid of values over which likelihood was computed over (ndarray)
    """
    # Apply prior for scale
    if 'scale' in options['priors_params'].keys():
        prior_scale = _define_distribution(grid['scale'], 
                                               options['priors_definitions']['scale'],
                                               options['priors_params']['scale'])
                
        # Log transform to prevent underflow 
        prior_scale = np.log(prior_scale)
        
        # Reshape to ensure broadcast works correctly
        prior_scale = prior_scale.reshape((-1,1,1,1))
        
        # Compute new posterior based on priors
        posterior = posterior + prior_scale
        
    # Apply prior for slope 
    if 'slope' in options['priors_params'].keys():
        prior_slope = _define_distribution(grid['slope'], 
                                              options['priors_definitions']['slope'],
                                              options['priors_params']['slope'])
        
        prior_slope = np.log(prior_slope)
        prior_slope = prior_slope.reshape((1,-1,1,1))
        posterior = posterior + prior_slope

    # Apply prior for gamma 
    if 'gamma' in options['priors_params'].keys():
        prior_gamma = _define_distribution(grid['gamma'], 
                                           options['priors_definitions']['gamma'],
                                           options['priors_params']['gamma'])

        prior_gamma = np.log(prior_gamma)
        prior_gamma = prior_gamma.reshape((1,1,-1,1))
        posterior = posterior + prior_gamma

    # Apply prior for lambda
    if 'lambda' in options['priors_params'].keys():
        prior_lambda = _define_distribution(grid['lambda'], 
                                            options['priors_definitions']['lambda'],
                                            options['priors_params']['lambda'])
    
        prior_lambda = np.log(prior_lambda)
        prior_lambda = prior_lambda.reshape((1,1,1,-1))
        posterior = posterior + prior_lambda
    
    return posterior


"""
*******************************************************
*
*  test_checkParams - UNIT TEST FOR TRAVIS CI
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   April 28, 2017
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES 
#################################################################
import numpy as np
from .checkParams import check_params as _check_params


#################################################################
#  DEFINE FUNCTIONS USED FOR UNIT TESTING
#################################################################

def _check_params_none(branch):    
    # Test cases for param_ests is None
    param_ests = None
    # Generate dataset 
    data01 = np.array([[0.0010,   45.0000,   90.0000],
                         [0.0015,   50.0000,   90.0000],
                         [0.0020,   44.0000,   90.0000],
                         [0.0025,   44.0000,   90.0000],
                         [0.0030,   52.0000,   90.0000],
                         [0.0035,   53.0000,   90.0000],
                         [0.0040,   62.0000,   90.0000],
                         [0.0045,   64.0000,   90.0000],
                         [0.0050,   76.0000,   90.0000],
                         [0.0060,   79.0000,   90.0000],
                         [0.0070,   88.0000,   90.0000],
                         [0.0080,   90.0000,   90.0000],
                         [0.0100,   90.0000,   90.0000]]);
    # Define nafc
    nafc = 2
    # Choose branch 
    if branch == 0:
        batch = False
        data = data01
    elif branch == 1:
        batch = True
        data = dict()
        data['data01'] = data01
    # Call function with arguments above
    _check_params(data, param_ests, nafc, batch, 'weibull')
    # Update success flag
    success = 1
    return success 

def _check_params_not_none(branch):    
    # Generate dataset 
    data = np.array([[0.0010,   45.0000,   90.0000],
                     [0.0015,   50.0000,   90.0000],
                     [0.0020,   44.0000,   90.0000],
                     [0.0025,   44.0000,   90.0000],
                     [0.0030,   52.0000,   90.0000],
                     [0.0035,   53.0000,   90.0000],
                     [0.0040,   62.0000,   90.0000],
                     [0.0045,   64.0000,   90.0000],
                     [0.0050,   76.0000,   90.0000],
                     [0.0060,   79.0000,   90.0000],
                     [0.0070,   88.0000,   90.0000],
                     [0.0080,   90.0000,   90.0000],
                     [0.0100,   90.0000,   90.0000]]);
    # Define nafc
    nafc = 2
    # Define batch
    batch = False
    # Choose branch 
    if branch == 0:
        # Set param_ests to trigger event
        param_ests = dict()
    elif branch == 1:
        # Set param_ests to trigger event
        param_ests = ['Trigger', 'Event']
    elif branch == 2:
        # Set param_ests to trigger event
        param_ests = [0.005]
    elif branch == 3:
        # Set param_ests vector with full length
        param_ests = [0.005, 2, None, 0]
    elif branch == 4:
        # Set param_ests to go through different number of args (trigger error)
        param_ests = [0.005, 2, 0.5, 0, 0.09]
    # Call function with arguments above
    _check_params(data, param_ests, nafc, batch, 'weibull')
    # Update success flag
    success = 1
    return success 

#################################################################
#  UNIT TESTS
#################################################################

def test_check_params_none_branch0():
    success = _check_params_none(0)
    # Assert if exception   
    assert success == 1

def test_check_params_none_branch1():
    success = _check_params_none(1)
    # Assert if exception   
    assert success == 1

def test_check_params_not_none_branch0():
    raised = False
    try:
        _check_params_not_none(0)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True

def test_check_params_not_none_branch1():
    raised = False
    try:
        _check_params_not_none(1)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_params_not_none_branch2():
    raised = False
    try:
        _check_params_not_none(2)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_params_not_none_branch3():
    success = _check_params_not_none(3)
    # Assert if exception   
    assert success == 1
    
def test_check_params_not_none_branch4():
    raised = False
    try:
        _check_params_not_none(6)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True



"""
*******************************************************
*
*  genGrid - GENERATE GRID FOR PRIORS AND POSTERIOR 
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   August 31, 2018
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
import scipy.stats as st


#################################################################
#  GENERATE GRID OF VALUES BASED ON FREE PARAMETERS TO ESTIMATE
#################################################################

def gen_grid(data, options):
    """Generates grid of values over which to compute the likelihood
    for each parameter to be estimated.
    
    Keyword arguments:
    data -- m x 3 numpy array
    options -- contains all options used to fit model (dictionary)
    """
    # Determine grid spacing and assignment for each parameter
    # Scale (typically alpha)
    if options['param_free'][0] is True:
        scale = np.linspace(data[:,0].min(), data[:,0].max(), options['density'])
    else:
        scale = options['param_ests'][0]
    
    # Slope (typically beta)
    if options['param_free'][1] is True:
        slope = np.linspace(options['param_ests'][1]*0.5, options['param_ests'][1]*1.5, options['density'])
    else: 
        slope = options['param_ests'][1]
    
    # Gamma (Guess rate)
    if options['param_free'][2] is True:
        lower_bound_gamma = options['param_ests'][2] - 0.25
        if lower_bound_gamma < 0:
            lower_bound_gamma = 0.0001
        upper_bound_gamma = options['param_ests'][2] + 0.25
        if upper_bound_gamma > 1:
            upper_bound_gamma = 0.9999
        gamma = np.linspace(lower_bound_gamma, upper_bound_gamma, options['density'])
    else:
        gamma = options['param_ests'][2]
        
    # Lambda (Lapse rate)
    if options['param_free'][3] is True:
        lambda_ = np.linspace(0.0001, 0.50, options['density'])
    else:
        lambda_ = options['param_ests'][3]

    # Determine meshgrid for posterior
    grid = dict()
    grid['A'], grid['B'], grid['G'], grid['L'] = np.meshgrid(scale, slope, gamma, lambda_)    
    
    # Save variables for defining and applying priors
    grid['scale'] = scale
    grid['slope'] = slope
    grid['gamma'] = gamma
    grid['lambda'] = lambda_
    
    # Output ditionary with grid
    return grid



"""
*******************************************************
*
*  plot_posterior - PLOT POSTERIOR
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   September 8, 2018
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
import matplotlib.pyplot as plt


#################################################################
#  PLOT POSTERIOR 
#################################################################
def plot_posterior(metrics):
    """Plots posterior surface for scale and slope parameters
    collapsing across guess and lapse rates.
    
    Keyword arguments:
    metrics -- contain important metrics about fitted model (dictionary)
    """
    # If other method other than numerical integration used, 
    # raise error 
    try: 
        metrics['posterior'] 
    except NameError: 
        raise ValueError('Posterior can only be plotted for numerical integration method (i.e., grid)')
    # Compute joint marginal posterior collapsing across nusiance 
    # parameters 
    posterior = np.sum(metrics['posterior'], axis = (2,3))
    # Generate plot of posterior with scale and 
    plt.contourf(metrics['Marginals_X']['scale'], 
                 metrics['Marginals_X']['slope'], 
                 posterior)
    plt.xlabel('scale', fontsize = 16)
    plt.ylabel('slope', fontsize = 16)
    plt.colorbar()
    plt.tight_layout()
    plt.show()



"""
*******************************************************
*
*  test_extract_priors - UNIT TEST FOR TRAVIS WITH EXTRACT PRIORS
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   May 29, 2018
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES 
#################################################################
import numpy as np
from .extractPriors import extract_priors as _extract_priors


#################################################################
#  DEFINE FUNCTIONS USED FOR UNIT TESTING
#################################################################

def _check_extract_priors(branch):    
    # Choose branch that defines sigmoid fit   
    if branch == 0:
        priors = ['Norm(0.5,2)']
    elif branch == 1:
        priors = ['Norm(0.5,2)', 'Beta(3,0.5)', 'Beta(3,0.5)', 'Beta(3,0.5)', 'Beta(3,0.5)']      
    elif branch == 2:
        priors = ['Norm(0.5,2)', 'Beta(3,0.5)', 'Norm(0.5,2)', 7]
    elif branch == 3:
        priors = ['Norm(0.5,2)', 'Beta(3,0.5)', 'Norm(0.5,2)', 'Beta(2,20)']
    # Use BayesFit to fit function
    options = dict()
    options['priors'] = priors
    _extract_priors(options)
    # Update success flag
    success = 1
    return success 


#################################################################
#  UNIT TESTS
#################################################################

def test_check_extract_priors_branch0():
    raised = False
    try:
        _check_extract_priors(0)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True

def test_check_extract_priors_branch1():
    raised = False
    try:
        _check_extract_priors(1)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_extract_priors_branch2():
    raised = False
    try:
        _check_extract_priors(2)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
def test_check_extract_priors_branch3():
    success = _check_extract_priors(3)
    # Assert if exception  
    assert success == 1



"""
*******************************************************
*
*  test_check_priors - UNIT TEST FOR TRAVIS WITH CHECK PRIORS
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   May 29, 2018
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES 
#################################################################
import numpy as np
from .extractPriors import extract_priors as _extract_priors
from .checkPriors import check_priors as _check_priors


#################################################################
#  DEFINE FUNCTIONS USED FOR UNIT TESTING
#################################################################

def _check_assigned_priors(branch):
    # Choose branch that defines sigmoid fit   
    if branch == 0:
        priors = ['Raise(0.5, 2)', 'Error(3,0.5)'] 
    elif branch == 1:
        priors = ['Norm(10150, 2)', 'Beta(3,0.5)']      
    # Check priors
    options = dict()
    options['priors'] = priors
    options = _extract_priors(options)
    _check_priors(options)
    # Update success flag
    success = 1
    return success


#################################################################
#  UNIT TESTS
#################################################################

def test_check_priors_branch0():
    raised = False
    try:
        _check_assigned_priors(0)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True

def test_check_priors_branch1():
    raised = False
    try:
        _check_assigned_priors(1)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True



"""
*******************************************************
*
*  get_Threshold - EXTRACT THRESHOLD FROM FITTED PARAMETERS
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   April 18, 2018
*  Last updated: September 12, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
from .psyFunction import psyfunction as _psyfunction


#################################################################
#  EXTRACT THRESHOLD FROM FITTED PARAMETERS
#################################################################
def get_threshold(data, options, metrics, threshold_pc, estimate_type = 'MAP'):
    """Check options provided ny user, or assigned by defaults
    are valid arguments.
    
    Keyword arguments:
    data -- m x 3 numpy array
    metrics -- contain important metrics about fitted model (dictionary)
    options -- contains all options used to fit model (dictionary)
    threshold_pc -- proportion correct to define threshold (float)
    estimate_type -- determines which metric to use for parameters of model (string)
    """
    # Extract threshold at specified level of propertion correct
    # using numerical approximation 
    def _find_nearest(y, x, value):
        idx = (np.abs(y-value)).argmin()
        return x[idx]
    
    x_est = np.linspace(data[:, 0].min(), data[:, 0].max(), int(1e4))
    
    if estimate_type not in ('MAP', 'Mean'):
        raise ValueError('Please provide appropriate argument for estimate type (e.g., MAP, Mean)')
        
    # Determine which values to use for vector of parameters 
    param_guess = np.zeros(4)
    counter = 0
    for keys in options['param_free']:
        if keys is True:
            param_guess[counter] = metrics[estimate_type][counter]
        elif keys is False: 
            param_guess[counter] = options['param_ests'][counter]
        counter += 1
    
    y_pred = _psyfunction(x_est, 
                          param_guess[0],
                          param_guess[1],
                          param_guess[2],
                          param_guess[3],
                          options['sigmoid_type'])
    
    threshold = _find_nearest(y_pred, x_est, threshold_pc)
        
    return threshold



"""
*******************************************************
*
*  calc_integral - APPROXIMATE INTEGRAL OF POSTERIOR
*
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   September 17, 2018
*  Last updated: September 17, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np


#################################################################
#  APPROXIMATE INTEGRAL USING NUMERICAL INTEGRATION
#################################################################

def calc_integral(posterior, options, metrics):
    """Compute integral of posterior surface using numerical
    integration.

    Keyword arguments:
    posterior -- 
    options -- 
    metrics --
    """
    
    # Generate list of free parameters to integrate
    list_free = []
    for i in range(0,4):
        if options['param_free'][i] is True:
            if i == 0:
                list_free.append('scale')
            elif i ==1:
                list_free.append('slope')
            elif i ==2:
                list_free.append('gamma') 
            elif i ==2:
                list_free.append('lambda')

    # Calculate step size and mass
    list_step = []
    for key in list_free:
        if metrics['Marginals_X'][key].shape[0] > 1:
            step = (metrics['Marginals_X'][key].max() - metrics['Marginals_X'][key].min()) / (len(metrics['Marginals_X'][key]) - 1)
            list_step.append(step)
    
    # Compute mass
    mass = np.product(list_step)
    
    # COmpute Integral
    integral = np.sum(posterior * mass)
    
    return integral


"""
*******************************************************
*
*  extractMetrics - EXTRACT RELEVANT METRICS FROM TRACE
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   September, 2017
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
from .psyFunction import psyfunction


#################################################################
#  COMPUTE MARGINAL DISTRIBUTION
#################################################################

# Compute marginal posterior distributions
def _compute_marginal(post_mat, axes_sum_over): 
    """Computes marginal distribution given a
    posterior surface.
    
    Keyword arguments:
    post_mat -- posterior surface (ndarray)
    axes_sum_over -- axes over which to sum across (tuple)
    """
    marginal = np.sum(post_mat, axes_sum_over) 
    return marginal


#################################################################
#  COMPUTE STANDARD DEVIATION
#################################################################
def _std_marginal(x_values, marginal_distro, mean_marginal_distro):
    """Computes standard deviation of marginal distribution.
    
    Keyword arguments:
    x_values -- stimulus intensities (ndarray)
    marginal_distro -- marginal distribution of specified parameter (ndarray)
    mean_marginal_distro -- expectation of marginal distribution (float)
    """
    return np.sqrt(np.sum( ((x_values - mean_marginal_distro)**2)*marginal_distro))


#################################################################
#  COMPUTE DEVIANCE OF MODEL
#################################################################
def deviance(data, options, metrics, estimate_type):
    """Computes deviance metric to assess
    goodness-of-fit of model.
    
    Keyword arguments:
    data -- m x 3 numpy array
    options -- contains all options used to fit model (dictionary)
    metrics -- contain important metrics about fitted model (dictionary)
    estimate_type -- determines which metric used for parameter estimates (string)
    """
    # Determine which values to use for vector of parameters 
    param_guess = np.zeros(4)
    counter = 0
    for keys in options['param_free']:
        if keys is True:
            param_guess[counter] = metrics[estimate_type][counter]
        elif keys is False: 
            param_guess[counter] = options['param_ests'][counter]
        counter += 1
    
    # Number of trials per intensity
    n = data[:,2]
    
    # Number of correct responses per intensity
    y = data[:,1]
    
    # First generate predictions based on user defined 
    # point estiamtes of function
    p = psyfunction(data[:,0], 
                      param_guess[0],
                      param_guess[1],
                      param_guess[2],
                      param_guess[3],
                      options['sigmoid_type'])
    p = np.rint(p*n)
    
    # Compute deviance measure for binomial model
    Deviance = 2*np.sum(n*y*np.log(y/p) + n*(1-y)*np.log((1-y)/(1-p)))
    
    return Deviance


#################################################################
#  COMPUTE PEARSON'S CHI SQUARED
#################################################################
def pearson_chisqr(data, options, metrics, estimate_type):
    """Computes Pearson's Chi-squared to assess
    goodness-of-fit of model.
    
    Keyword arguments:
    data -- m x 3 numpy array
    options -- contains all options used to fit model (dictionary)
    metrics -- contain important metrics about fitted model (dictionary)
    estimate_type -- determines which metric used for parameter estimates (string)
    """
    # Determine which values to use for vector of parameters 
    param_guess = np.zeros(4)
    counter = 0
    for keys in options['param_free']:
        if keys is True:
            param_guess[counter] = metrics[estimate_type][counter]
        elif keys is False: 
            param_guess[counter] = options['param_ests'][counter]
        counter += 1
    
    # Number of trials per intensity
    n = data[:,2]
    
    # Number of correct responses per intensity
    y = data[:,1]
    
    # First generate predictions based on user defined 
    # point estiamtes of function
    p = psyfunction(data[:,0], 
                      param_guess[0],
                      param_guess[1],
                      param_guess[2],
                      param_guess[3],
                      options['sigmoid_type'])
    p = np.rint(p*n)
    
    # Compute Pearson's Chi Squared 
    p_chisqr = np.sum(((y-p)**2)/p)
    
    return p_chisqr
    
    
#################################################################
#  COMPUTE HIGHEST POSTERIOR DENSITY INTERVAL
#################################################################

def hpdi(marginal, marginal_labels, alpha):
    """Computes Highest Posterior Density Interval (HPDI)
    for a marginal distribution given a specified level
    of confidence.
    
    Keyword arguments:
    marginal -- marginal distribution to compute bci (ndarray)
    marginal_labels -- values over which marginal was computed (ndarray)
    alpha -- confidence interval (float)
    """
    # Calculate credible mass
    cred_mass = 1 - alpha
    
    # Sort marginal distribution
    # Sort marginal and extract indices
    sort_idx = np.argsort(marginal)
    # Sort marginal vector itself
    sort_marg = np.sort(marginal)[::-1]
    # Label Idxs as binary 
    binary_idx = np.where(np.cumsum(sort_marg) <= cred_mass)

    # Label points that are part of highest posterior interval
    hpdi = marginal_labels[sort_idx[binary_idx]]
    
    # Calculate cutoff probability
    # i.e., above this probability, all points are part of hpdi
    hpdi_px = sort_marg[binary_idx].min()

    return hpdi_px, hpdi


#################################################################
#  COMPUTE BAYESIAN CREDIBLE INTERVAL
#################################################################

# Compute Bayesian Credible Interval   
def bci(marginal, alpha):
    """Computes Bayesian Credible Interval (BCI) for
    a marginal distribution given a specified level
    of confidence.
    
    Keyword arguments:
    marginal -- marginal distribution to compute bci (ndarray)
    alpha -- confidence interval (float)
    """
    # Compute cumulative sum of marginal distribution
    cpxx = np.cumsum(marginal) 
    
    # Boundary
    bound = (1-alpha)/2
    # Lower boundary 
    lower = marginal[np.where(cpxx >= 0 + bound)][0]   
    # Upper boundary
    upper = marginal[np.where(cpxx >= 1 - bound)][0]   
    
    return lower, upper 


#################################################################
#  EXTRACT RELEVANT PARAMETERS FROM POSTERIOR GRID
#################################################################
# Extract fitted parameters and other metrics 
# of importance from posterior grid
def extract_metrics_grid(data, options, metrics, posterior, grid):
    """Computes relevant metrics from fitted model.
    
    Keyword arguments:
    data -- m x 3 numpy array
    options -- contains all options used to fit model (dictionary)
    metrics -- contain important metrics about fitted model (dictionary)
    posterior -- posterior surface (ndarray)
    grid -- grid of values over which likelihood was computed over (ndarray) 
    """
    
    #################################################################
    #  COMPUTE MARGINAL POSTERIOR DISTRIBUTIONS
    #################################################################
    
    # Compute marginal posterior distributions
    metrics['Marginals'] = dict()
    metrics['Marginals']['scale'] = _compute_marginal(posterior, (1,2,3))
    metrics['Marginals']['slope'] = _compute_marginal(posterior, (0,2,3))
    metrics['Marginals']['gamma'] = _compute_marginal(posterior, (0,1,3))
    metrics['Marginals']['lambda'] = _compute_marginal(posterior, (0,1,2))
    
    #################################################################
    #  GET MAP ESTIMATE FROM POSTERIOR 
    #################################################################
    
    # Get index of MAP estimate
    map_index = np.where(posterior == np.max(posterior.ravel()))
    
    # Form vector for MAP estimate of each parameter if free parameter
    metrics['MAP'] = np.zeros(4)
    for i in range(0,4):
        if options['param_free'][i] is True:
            if i == 0:
                metrics['MAP'][i] = grid['A'][map_index]
            elif i == 1: 
                metrics['MAP'][i] = grid['B'][map_index]
            elif i == 2: 
                metrics['MAP'][i] = grid['G'][map_index]   
            elif i == 3: 
                metrics['MAP'][i] = grid['L'][map_index]
        elif options['param_free'][i] is False:
            metrics['MAP'][i] = np.nan
    
    #################################################################
    #  GET MEAN ESTIMATE FROM MARGINALS 
    #################################################################
    
    # Form vector for MAP estimate of each parameter if free parameter
    metrics['Mean'] = np.zeros(4)
    for i in range(0,4):
        if options['param_free'][i] is True:
            if i == 0:
                metrics['Mean'][i] = np.sum(metrics['Marginals']['scale']*grid['scale']) 
            elif i == 1:                 
                metrics['Mean'][i] = np.sum(metrics['Marginals']['slope']*grid['slope']) 
            elif i == 2:                 
                metrics['Mean'][i] = np.sum(metrics['Marginals']['gamma']*grid['gamma']) 
            elif i == 3:                 
                metrics['Mean'][i] = np.sum(metrics['Marginals']['lambda']*grid['lambda']) 
        elif options['param_free'][i] is False:
            metrics['Mean'][i] = np.nan
    
    #################################################################
    #  CALCULATE SD FOR EACH PARAMETER
    #################################################################
    
    # Form vector for MAP estimate of each parameter if free parameter
    metrics['SD'] = np.zeros(4)
    for i in range(0,4):
        if options['param_free'][i] is True:
            if i == 0:
                metrics['SD'][i] = _std_marginal(grid['scale'], 
                                                   metrics['Marginals']['scale'], 
                                                   metrics['Mean'][0])
            elif i == 1:                 
                metrics['SD'][i] = _std_marginal(grid['slope'], 
                                                   metrics['Marginals']['slope'], 
                                                   metrics['Mean'][1])
            elif i == 2:                 
                metrics['SD'][i] = _std_marginal(grid['gamma'], 
                                                   metrics['Marginals']['gamma'], 
                                                   metrics['Mean'][2])
            elif i == 3:                 
                metrics['SD'][i] = _std_marginal(grid['lambda'], 
                                                   metrics['Marginals']['lambda'], 
                                                   metrics['Mean'][3])
        elif options['param_free'][i] is False:
            metrics['SD'][i] = np.nan

    #################################################################
    #  CALCULATE BOUNDS FOR HIGHEST POSTERIOR DENSITY
    #################################################################
    
    # Form vector for MAP estimate of each parameter if free parameter
    metrics['BCI'] = np.zeros(shape=(2,4))
    for i in range(0,4):
        if options['param_free'][i] is True:
            if i == 0:
                metrics['BCI'][0,i], metrics['BCI'][1,i] = bci(metrics['Marginals']['scale'], 0.05)
            elif i == 1:                 
                metrics['BCI'][0,i], metrics['BCI'][1,i] = bci(metrics['Marginals']['slope'], 0.05)
            elif i == 2:                 
                metrics['BCI'][0,i], metrics['BCI'][1,i] = bci(metrics['Marginals']['gamma'], 0.05)
            elif i == 3:                 
                metrics['BCI'][0,i], metrics['BCI'][1,i] = bci(metrics['Marginals']['lambda'], 0.05)
        elif options['param_free'][i] is False:
            metrics['BCI'][0,i] = np.nan
            metrics['BCI'][1,i] = np.nan
            
    #################################################################
    #  CALCULATE DEVIANCE
    #################################################################
    metrics['deviance'] = deviance(data, options, metrics, 'MAP')
    
    #################################################################
    #  CALCULATE DEVIANCE
    #################################################################
    metrics['pearsons_chisqr'] = pearson_chisqr(data, options, metrics, 'MAP')
    
    return metrics



"""
*******************************************************
*
*  checkOptions - CHECK USER PROVIDED ARGS FOR OPTIONS
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   April 14, 2018
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np


#################################################################
#  CHECK OPTIONS PROVIDED
#################################################################
def check_options(options):
    """Check options provided ny user, or assigned by defaults
    are valid arguments.
    
    Keyword arguments:
    options -- contains all options used to fit model (dictionary)
    """
    if options['batch'] not in (True, False):
        raise ValueError('User Error: Invalid argument provided for batch type.')
    if options['sigmoid_type'] not in ('norm', 'logistic', 'gumbel', 'weibull', 'quick', 'log-quick', 'hyperbolic'):
        raise ValueError('User Error: Sigmoid type specified is not one made available by the module.')
    if options['logspace'] not in (True, False):
        raise ValueError('User/Internal Error: Invalid argument for log-spacing of x-values provided.')
    if options['nafc'] not in (1,2,3,4,5,6,7,8,9,10):
        raise ValueError('Warning: Value provided for nAFC is greater than 10, or less than zero!')
    if isinstance(options['threshold'], (int, float, complex)) is False:
        raise ValueError('Please provide a numerical argument for proportion correct estimate of threshold.')
    if not 0 <= options['threshold'] <= 1:
        raise ValueError('User Error: Value provided for threshold estimate not between zero and one!')
    if isinstance(options['nafc'], int) is False:
        raise ValueError('User Error: Please provide a integer for nAFC argument.')
    if isinstance(options['param_ests'], list) is False:
        raise ValueError('User Error: Please provide a argument of type list for the parameter estimates of the model.')
    if isinstance(options['param_free'], list) is False:
        raise ValueError('''User Error: Please provide a argument of type list for the 
                           parameter constraints of the model.''')
    if isinstance(options['density'], int) is False:
        raise ValueError('User Error: Please provide a integer for density of grid.')


"""
*******************************************************
*
*  gridLikelihood - COMPUTES LIKELIHOOD USING GRID EVALUATION
*
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   September 3, 2018
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
from .psyFunction import psyfunction as _psyfunction


#################################################################
#  COMPUTE LIKELIHOOD USING GRID EVALUATION
#################################################################

def grid_likelihood(data, options, grid):
    """Computes likelihood across of parameters defined by grid
    given the data. 
    
    Keyword arguments:
    data -- m x 3 numpy array
    options -- contains all options used to fit model (dictionary)
    grid -- grid of values over which to compute likelihood over (ndarray) 
    """
    # Initialize posterior
    likelihood = np.zeros(shape = grid['A'].shape)
    
    # Compute likelihood
    for i in range(0,len(data[:,0])):
        y = _psyfunction(data[i,0], 
                         grid['A'], 
                         grid['B'], 
                         grid['G'], 
                         grid['L'], 
                         options['sigmoid_type'])*.99 + 0.005
                         
        likelihood += data[i,1]*np.log(y) + (data[i,2] - data[i,1])*np.log(1-y)
        
    return likelihood



"""
*******************************************************
*
*  plot_priors - PLOT PRIORS
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   September 10, 2018
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
import scipy.stats as st
import matplotlib.pyplot as plt


#################################################################
#  FUNCTION THAT DEFINES DISTRIBUTION BASED ON CLASS
#################################################################

def _define_distribution(xx, user_defined_distro, distro_parameters):
    """Assigns probability density to parameters defined by grid
    and specified distribution.
    
    Keyword arguments:
    xx -- range of values over which to define distribution (ndarray)
    user_defined_distro -- name of distribution (string)
    distro_parameters -- parameters for distribution (list)
    """
    # Define distrobution based on inputs
    if user_defined_distro == 'Unif':
        distro = st.uniform.pdf(xx, 
                                distro_parameters[0], 
                                distro_parameters[1])
    elif user_defined_distro == 'Norm':
        distro = st.norm.pdf(xx, 
                             distro_parameters[0], 
                             distro_parameters[1])
    elif user_defined_distro == 'Log-Norm':
        distro = st.lognorm.pdf(xx, 
                                distro_parameters[0], 
                                distro_parameters[1])
    elif user_defined_distro == 'Beta':
        distro = st.beta.pdf(xx, 
                             distro_parameters[0], 
                             distro_parameters[1])
    elif user_defined_distro == 'Gamma':
        distro = st.gamma.pdf(xx, 
                              distro_parameters[0], 
                              distro_parameters[1])
    
    # Ensure distribution sums to one
    distro /= distro.sum()
        
    return distro


#################################################################
#  PLOT PRIORS
#################################################################
def plot_priors(options, metrics):
    """Plots prior distribution defined by user when fitting model.
    
    Keyword arguments:
    options -- contains all options used to fit model (dictionary)
    metrics -- contain important metrics about fitted model (dictionary)
    """
    # Generate basic plot of marginal distributions
    fig, axes = plt.subplots(2, 2, 
                             subplot_kw=dict(polar=False), 
                             figsize = (7,6))
    # Label axes
    axes[0,0].set_xlabel('Scale')
    axes[0,1].set_xlabel('Slope')
    axes[1,0].set_xlabel('Gamma')
    axes[1,1].set_xlabel('Lambda')
    axes[0,0].set_ylabel('Probability')
    axes[0,1].set_ylabel('Probability')
    axes[1,0].set_ylabel('Probability')
    axes[1,1].set_ylabel('Probability')
    # Scale parameter
    if 'scale' in options['priors_params'].keys():
        prior = _define_distribution(metrics['Marginals_X']['scale'], 
                                       options['priors_definitions']['scale'],
                                       options['priors_params']['scale'])
        axes[0,0].plot(metrics['Marginals_X']['scale'],
            prior, 
            lw=3, 
            color='#5998ff')
        axes[0,0].fill_between(metrics['Marginals_X']['scale'],
            prior, color='#5998ff', alpha = .4)
    else:
        axes[0,0].text(0.5,0.5, "None",
            horizontalalignment='center', 
            verticalalignment='center', 
            transform=axes[0,0].transAxes)       
    # Slope parameter
    if 'slope' in options['priors_params'].keys():
        prior = _define_distribution(metrics['Marginals_X']['slope'], 
                                       options['priors_definitions']['slope'],
                                       options['priors_params']['slope'])
        axes[0,1].plot(metrics['Marginals_X']['slope'],
            prior, 
            lw=3, 
            color='#5998ff')
        axes[0,1].fill_between(metrics['Marginals_X']['slope'],
            prior, color='#5998ff', alpha = .4)
    else:
        axes[0,1].text(0.5,0.5, "None",
            horizontalalignment='center', 
            verticalalignment='center', 
            transform=axes[0,1].transAxes)     
    # Gamma 
    if 'gamma' in options['priors_params'].keys():
        prior = _define_distribution(metrics['Marginals_X']['gamma'], 
                                       options['priors_definitions']['gamma'],
                                       options['priors_params']['gamma'])
        axes[1,0].plot(metrics['Marginals_X']['gamma'],
            prior, 
            lw=3, 
            color='#5998ff')
        axes[1,0].fill_between(metrics['Marginals_X']['gamma'],
            prior, color='#5998ff', alpha = .4)
    else:
        axes[1,0].text(0.5,0.5, "None",
            horizontalalignment='center', 
            verticalalignment='center', 
            transform=axes[1,0].transAxes)     
    # Lambda
    if 'lambda' in options['priors_params'].keys():
        prior = _define_distribution(metrics['Marginals_X']['lambda'], 
                                       options['priors_definitions']['lambda'],
                                       options['priors_params']['lambda'])
        axes[1,1].plot(metrics['Marginals_X']['lambda'],
            prior, 
            lw=3, 
            color='#5998ff')
        axes[1,1].fill_between(metrics['Marginals_X']['lambda'],
            prior, color='#5998ff', alpha = .4)
    else:
        axes[1,1].text(0.5,0.5, "None",
            horizontalalignment='center', 
            verticalalignment='center', 
            transform=axes[1,1].transAxes)     
    
    plt.tight_layout()
    plt.show()



"""
*******************************************************
*
*  checkLogspace - LOGSPACE INTENSITIES IF NEEDED
*
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   September 2017
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np


#################################################################
#  CONVERT X-VALUES IN FIRST COLUMN TO LOG-SPACED VALUES
#################################################################
# Convert x-values to logspace
def convert_logspace(data):
    """Log-transforms stimulus intensities.

    Keyword arguments:
    data -- m x 3 numpy array
    """
    data_x_logtrans = np.log10(data[:, 0])
    data_trans = np.array([data_x_logtrans, data[:,1], data[:,2]]).T
    return data_trans


#################################################################
#  CHECK WHETHER INTENSITIES NEED TO BE LOGSPACED
################################################################
# Check sigmoid type provided to convert x-values to logspace if required
def check_logspace(data, logspace, sigmoid_type, batch):
    """Performs a check of whether data need to be logspaced
    based on user input via logspace option and sigmoid type
    chosen (i.e, Weibull).

    Keyword arguments:
    data -- m x 3 numpy array
    logspace -- specifies whether to logspace data (logical)
    sigmoid_type -- sigmoid type fit to data (string)
    """
    data_copy = data
    # Ensure argument provided for logspace is valid
    if logspace not in (None, True, False):
        raise Exception('User/Internal Error: Invalid argument for log-spacing of x-values provided.')
    # Run through options and logspace x-values if needed
    if logspace is None:
        if sigmoid_type in ('weibull', 'log-quick'):
            if batch is False:
                logspace = True
                if data[:, 0].min() < 0:
                    raise Exception('Cannot convert negative data points to log-spaced values!')
                # Convert x-values in data
                data_copy = convert_logspace(data)
            elif batch is True:
                logspace = True
                for keys in data:
                    if data[keys][:, 0].min() < 0:
                        raise Exception('Cannot convert negative data points to log-spaced values!')
                    # Convert x-values in data
                    data_copy[keys] = convert_logspace(data[keys])
        else:
            logspace = False
            data_copy = data
    elif logspace is True:
        if batch is False:
            logspace = True
            if data[:, 0].min() < 0:
                raise Exception('Cannot convert negative data points to log-spaced values!')
            # Convert x-values in data
            data_copy = convert_logspace(data)
        elif batch is True:
            logspace = True
            for keys in data:
                if data[keys][:, 0].min() < 0:
                    raise Exception('Cannot convert negative data points to log-spaced values!')
                # Convert x-values in data
                data_copy[keys] = convert_logspace(data[keys])
    elif logspace is False:
        data_copy = data
    return data_copy, logspace



"""
*******************************************************
*
*  checkData - CHECK USER PROVIDED DATA
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   April 14, 2018
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np


#################################################################
#  CHECK DATA STRUCTURE
#################################################################
# Check data structure provided by user
# Format requested is a N x 3 data frame such that [x, y, N]
def check_data(data, batch):
    """Performs a check of the data provided by the user
    to ensure to meets the formatting requirements used by
    Bayesfit

    Keyword arguments:
    data -- m x 3 numpy array
    batch -- specifies whether batch fitting data (logical)
    """
    if batch is False:
        if isinstance(data, np.ndarray) is False:
            raise Exception('Single dataset provided is not of type: np.ndarray. Please change data type provided.') 
        elif isinstance(data, np.ndarray) is True:
            if data.shape[1] != 3:
                raise Exception('Data provided does not contain the number of columns required! (i.e., [x, y, N])')
            for n in data[:,2]:
                if n < 1:
                    raise ValueError('''The number of trials for a stimulus intensity is less than one!''')
    elif batch is True:
        if isinstance(data, dict) is False:
            raise Exception('Batch dataset provided is not of type: dictionary. Please change data type provided.')
        elif isinstance(data, dict) is True:
            for keys in data:
                if data[keys].shape[1] != 3:
                    raise Exception('''A dataset provided in batch does not contain the number of columns required! 
                     (i.e., [x, y, N])''')
                for n in data[keys][:,2]:
                    if n < 1:
                        raise ValueError('''The number of trials for a stimulus intensity is less than one!''')



"""
*******************************************************
*
*  checkPriors - CHECK PRIORS PROVIDED BY USER 
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   May 19, 2018
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np


#################################################################
#  CHECK PRIORS PROVIDED
#################################################################
# Function to check that priors provided by the user are valid 
def check_priors(options): 
    """Checks whether priors provided by user, and default
    values assigned by extract_metrics function are valid 
    assignments.

    Keyword arguments:
    options -- contains all options used to fit model (dictionary)
    """
    # Check whether distribution types provided are valid
    for keys in options['priors_definitions']:
        if options['priors_definitions'][keys] not in ('Unif', 'Norm', 'Log-Norm', 'Beta', 'Gamma'): 
            raise Exception('''One of the prior distributions defined is not 
                            a valid argument from options available!''')
    # Check whether parameters for distributions within acceptable bounds
    # Note: Does NOT check whether parameters for each distribution are valid.
    for values in options['priors_params']:
        for i in range(0,len(options['priors_params'][values])):
            if options['priors_params'][values][i] > 10000 or options['priors_params'][values][i] < -10000:
                    raise Exception('''One of the parameters that define a prior
                    distribution is greater than 10000 or less than -10000!''')


"""
*******************************************************
*
*  plot_psyFcn - PLOT MODEL FIT TO DATA
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   April 18, 2017
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
import matplotlib.pyplot as plt
from .psyFunction import psyfunction as _psyfunction
from .checkLogspace import check_logspace as _check_logspace


#################################################################
#  GENERATE PLOT OF MODEL FIT TO DATA
#################################################################
def plot_psyfcn(data, 
                options,
                metrics,
                log_xscale = False, 
                scale_pnts = True,
                estimate_type = 'MAP'):
    """Plots psychometric function from fitted model.
    
    Keyword arguments:
    data -- m x 3 numpy array
    options -- contains all options used to fit model (dictionary)
    metrics -- contain important metrics about fitted model (dictionary)
    log_xscale -- log-transform x-scale (logical)
    scale_pnts -- scale size of points by N-trial number (logical)
    estimate_type -- determines which metric to use for parameters of model (string)
    """
    
    if estimate_type not in ('MAP', 'Mean'):
        raise ValueError('Please provide appropriate argument for estimate type (e.g., MAP, Mean)')
    
    # Determine which values to use for vector of parameters 
    param_guess = np.zeros(4)
    counter = 0
    for keys in options['param_free']:
        if keys is True:
            param_guess[counter] = metrics[estimate_type][counter]
        elif keys is False: 
            param_guess[counter] = options['param_ests'][counter]
        counter += 1

    # Generate copy of data
    data_copy = data
    # Check whether x-values need to be log-spaced
    data_copy, logspace = _check_logspace(data, 
                                          options['logspace'], 
                                          options['sigmoid_type'],
                                          options['batch'])
    
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(5, 5))
    
    # Generate smooth curve from fitted function
    x_max = data_copy[:, 0].max()
    x_min = data_copy[:, 0].min()
    x_est = np.linspace(x_min, x_max, 1000)
    
    y_pred = _psyfunction(x_est, 
                          param_guess[0],
                          param_guess[1],
                          param_guess[2],
                          param_guess[3],
                          options['sigmoid_type'])

    if log_xscale is True:
        ax.set_xscale('log')
    ax.plot([x_min, metrics['threshold']],
            [options['threshold'], options['threshold']],
            color='black',
            linestyle='dotted',
            linewidth=2,
            zorder=1)
    ax.plot([metrics['threshold'], metrics['threshold']], 
            [0, options['threshold']],
            color='black',
            linestyle='dotted',
            linewidth=2,
            zorder=1)
    ax.set_ylim(-0.05, 1.05)
    ax.axhline(y=param_guess[2], color='r', linestyle='dashed', linewidth=2, zorder=1)
    # Plot remainder of curve            
    if scale_pnts is True:
        for i in range(data[:,0].shape[0]): 
            ax.scatter(data_copy[i,0], 
                       data_copy[i,1]/data_copy[i,2], 
                       color='#5998ff', 
                       s=data_copy[i,2]*2, 
                       alpha=0.5, 
                       zorder=5, 
                       marker='o')
    elif scale_pnts is False: 
        ax.scatter(data_copy[:,0], 
                   data_copy[:,1]/data_copy[:,2], 
                   color='#5998ff', 
                   s=135, 
                   alpha=0.5, 
                   zorder=5, 
                   marker='o')    
    ax.plot(x_est, y_pred, linestyle='-', color='black', linewidth=2, alpha=0.85, zorder=10)
    ax.set_xlabel('Stimulus Intensity', fontsize=16)
    ax.set_ylabel('Proportion correct', fontsize=16)
    ax.xaxis.set_tick_params(labelsize=13)
    ax.yaxis.set_tick_params(labelsize=13)
    plt.tight_layout()
    plt.show()



"""
*******************************************************
*
*  test_bayesfit - UNIT TEST FOR TRAVIS CI
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   April 28, 2017
*  Last updated: September 17, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES 
#################################################################
import numpy as np
import bayesfit as bf


#################################################################
#  DEFINE FUNCTIONS USED FOR UNIT TESTING
#################################################################

def _check_bayesfit_psyfunctions(branch):    
    # Generate dataset 
    data = np.array([[0.0010,   45.0000,   90.0000],
                     [0.0015,   50.0000,   90.0000],
                     [0.0020,   44.0000,   90.0000],
                     [0.0025,   44.0000,   90.0000],
                     [0.0030,   52.0000,   90.0000],
                     [0.0035,   53.0000,   90.0000],
                     [0.0040,   62.0000,   90.0000],
                     [0.0045,   64.0000,   90.0000],
                     [0.0050,   76.0000,   90.0000],
                     [0.0060,   79.0000,   90.0000],
                     [0.0070,   88.0000,   90.0000],
                     [0.0080,   90.0000,   90.0000],
                     [0.0100,   90.0000,   90.0000]]);
    # Choose branch that defines sigmoid fit   
    if branch == 0:
        sigmoid_type = 'logistic'
    elif branch == 1:
        sigmoid_type = 'weibull'        
    elif branch == 2:
        sigmoid_type = 'gumbel'
    elif branch == 3:
        sigmoid_type = 'quick'
    elif branch == 4:
        sigmoid_type = 'log-quick'
    elif branch == 5:
        sigmoid_type = 'hyperbolic'
    elif branch == 6:
        sigmoid_type = 'norm'
    # Use BayesFit to fit function
    metrics, options = bf.fitmodel(data, sigmoid_type = sigmoid_type, logspace=False)
    # Update success flag
    success = 1
    return success 

def _check_bayesfit_param_constraints(branch):    
    # Generate dataset 
    data = np.array([[0.0010,   45.0000,   90.0000],
                     [0.0015,   50.0000,   90.0000],
                     [0.0020,   44.0000,   90.0000],
                     [0.0025,   44.0000,   90.0000],
                     [0.0030,   52.0000,   90.0000],
                     [0.0035,   53.0000,   90.0000],
                     [0.0040,   62.0000,   90.0000],
                     [0.0045,   64.0000,   90.0000],
                     [0.0050,   76.0000,   90.0000],
                     [0.0060,   79.0000,   90.0000],
                     [0.0070,   88.0000,   90.0000],
                     [0.0080,   90.0000,   90.0000],
                     [0.0100,   90.0000,   90.0000]]);
    # Choose branch that defines sigmoid fit   
    if branch == 0:
        param_free = [True, True, False, False]
    elif branch == 1:
        param_free = [True, True, True, False]      
    elif branch == 2:
        param_free = [True, True, True, True]
    # Use BayesFit to fit function
    metrics, options = bf.fitmodel(data, param_free = param_free, density = 20)
    # Update success flag
    success = 1
    return success 

def _check_bayesfit_priors():    
    # Generate dataset 
    data = np.array([[0.0010,   45.0000,   90.0000],
                     [0.0015,   50.0000,   90.0000],
                     [0.0020,   44.0000,   90.0000],
                     [0.0025,   44.0000,   90.0000],
                     [0.0030,   52.0000,   90.0000],
                     [0.0035,   53.0000,   90.0000],
                     [0.0040,   62.0000,   90.0000],
                     [0.0045,   64.0000,   90.0000],
                     [0.0050,   76.0000,   90.0000],
                     [0.0060,   79.0000,   90.0000],
                     [0.0070,   88.0000,   90.0000],
                     [0.0080,   90.0000,   90.0000],
                     [0.0100,   90.0000,   90.0000]]);
    priors = ['Norm(0.005,0.001)', 'Norm(3,0.75)', None, None]
    metrics, options = bf.fitmodel(data, priors = priors, sigmoid_type='weibull', logspace=False)
    # Update success flag
    success = 1
    return success 

def _check_bayesfit_batch_process(branch):    
    # Generate dataset 
    data_batch = np.array([[0.0010,   45.0000,   90.0000],
		                     [0.0015,   50.0000,   90.0000],
		                     [0.0020,   44.0000,   90.0000],
		                     [0.0025,   44.0000,   90.0000],
		                     [0.0030,   52.0000,   90.0000],
		                     [0.0035,   53.0000,   90.0000],
		                     [0.0040,   62.0000,   90.0000],
		                     [0.0045,   64.0000,   90.0000],
		                     [0.0050,   76.0000,   90.0000],
		                     [0.0060,   79.0000,   90.0000],
		                     [0.0070,   88.0000,   90.0000],
		                     [0.0080,   90.0000,   90.0000],
		                     [0.0100,   90.0000,   90.0000]]);
    # Create batch data object 
    data = dict()
    data['data01'] = data_batch
    data['data02'] = data_batch
    # Choose branch that defines sigmoid fit   
    if branch == 0:
        param_free = [True, True, False, False]
    elif branch == 1:
        param_free = [True, True, True, False]      
    elif branch == 2:
        param_free = [True, True, True, True]
    # Use BayesFit to fit function
    metrics, options = bf.fitmodel(data, batch = True, param_free = param_free, density = 20)
    # Update success flag
    success = 1
    return success 

#################################################################
#  UNIT TESTS
#################################################################

def test_check_bayesfit_psyfunctions_branch0():
    success = _check_bayesfit_psyfunctions(0)
    # Assert if exception  
    assert success == 1

def test_check_bayesfit_psyfunctions_branch1():
    success = _check_bayesfit_psyfunctions(1)
    # Assert if exception  
    assert success == 1
    
def test_check_bayesfit_psyfunctions_branch2():
    success = _check_bayesfit_psyfunctions(2)
    # Assert if exception  
    assert success == 1
    
def test_check_bayesfit_psyfunctions_branch3():
    success = _check_bayesfit_psyfunctions(3)
    # Assert if exception  
    assert success == 1
    
def test_check_bayesfit_psyfunctions_branch4():
    success = _check_bayesfit_psyfunctions(4)
    # Assert if exception  
    assert success == 1

def test_check_bayesfit_psyfunctions_branch5():
    success = _check_bayesfit_psyfunctions(5)
    # Assert if exception  
    assert success == 1  

def test_check_bayesfit_psyfunctions_branch6():
    success = _check_bayesfit_psyfunctions(6)
    # Assert if exception  
    assert success == 1  
    
def test_check_bayesfit_param_constraints_branch0():
    success = _check_bayesfit_param_constraints(0)
    # Assert if exception  
    assert success == 1    
    
def test_check_bayesfit_param_constraints_branch1():
    success = _check_bayesfit_param_constraints(1)
    # Assert if exception  
    assert success == 1      
    
def test_check_bayesfit_param_constraints_branch2():
    success = _check_bayesfit_param_constraints(2)
    # Assert if exception  
    assert success == 1     

def test_check_bayesfit_batch_process_branch0():
    success = _check_bayesfit_batch_process(0)
    # Assert if exception  
    assert success == 1    

def test_check_bayesfit_batch_process_branch1():
    success = _check_bayesfit_batch_process(1)
    # Assert if exception  
    assert success == 1   

def test_check_bayesfit_batch_process_branch2():
    success = _check_bayesfit_batch_process(2)
    # Assert if exception  
    assert success == 1   

def test_check_bayesfit_priors():
    success = _check_bayesfit_priors()
    # Assert if exception  
    assert success == 1   



"""
*******************************************************
*
*  psyFunction - GENERATE GENERAL FORMULATION OF PSYCHOMETRIC FUNCTION
*
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   April 18, 2018
*  Last updated: August 30, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
from scipy.special import erf


#################################################################
#  DEFINE PSYCHOMETRIC FUNCTIONS
#################################################################
# Cumulative Gaussian
def _norm(x, xShift, sd):
    return ((erf((x-xShift)/(np.sqrt(2)*sd))+1)*0.5)

# Logistic
def _logistic(x, alpha, beta):
    return 1 / (1 + np.exp(-beta*(x - alpha)))


# Weibull
def _weibull(x, alpha, beta):
    return 1 - np.exp(-(x/alpha)**beta) 


# Gumbel
def _gumbel(x, alpha, beta):
    return 1 - np.exp(-10**(beta*(x-alpha)))


# Quick
def _quick(x, alpha, beta):
    return 1 - 2**(-(x/alpha)**beta)


# Log-Quick
def _quicklog(x, alpha, beta):
    return 1 - 2**(-10**(beta*(x - alpha)))


# Hyperbolic Secant
def _hyperbolic_secant(x, alpha, beta):
    return (2/np.pi)*np.arctan(np.exp((np.pi/2)*beta*(x - alpha)))


#################################################################
#  DEFINE GENERIC FORMULATION OP PSYCHOMETRIC FUNCTIONS
#################################################################

# Note: Although laballed as scale and slope, these parameters
# can jointly determine such characteristics depending upon
# the PF function chosen (e.g., Weibull).
def psyfunction(x, scale, slope, gamma, lambda_, sigmoid_type):
    with np.errstate(divide='ignore',invalid='ignore', over ='ignore'):
        if sigmoid_type == 'norm':
            return gamma + (1 - lambda_ - gamma) * _norm(x, scale, slope)
        elif sigmoid_type == 'logistic':
            return gamma + (1 - lambda_ - gamma) * _logistic(x, scale, slope)
        elif sigmoid_type == 'weibull':
            return gamma + (1 - lambda_ - gamma) * _weibull(x, scale, slope)
        elif sigmoid_type == 'gumbel':
            return gamma + (1 - lambda_ - gamma) * _gumbel(x, scale, slope)
        elif sigmoid_type == 'quick':
            return gamma + (1 - lambda_ - gamma) * _quick(x, scale, slope)
        elif sigmoid_type == 'log-quick':
            return gamma + (1 - lambda_ - gamma) * _quicklog(x, scale, slope)
        elif sigmoid_type == 'hyperbolic':
            return gamma + (1 - lambda_ - gamma) * _hyperbolic_secant(x, scale, slope)



"""
*******************************************************
*
*  checkParams - CHECK USER PROVIDED ESTIMATES FOR PARAMETERS
*
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   April 14, 2018
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES
#################################################################
import numpy as np
from .psyFunction import psyfunction as _psyfunction
from scipy.optimize import curve_fit
import warnings


#################################################################
#  COMPUTE ESTIMATES FOR SCALE/SLOPE PARAMETERS
#################################################################
def _param_guess(data, nafc, sigmoid_type):
    """Generate initial estimate for parameters governing 
    scale and slope of sigmoidal function.

    Keyword arguments:
    data -- m x 3 numpy array
    nafc -- N-alternative forced choice experiment (integer)
    sigmoid_type -- sigmoid type fit to data (string)
    """
    # Wrapper function to avoid errors with passing sigmoid type defintion
    # to curve fit function
    def wrapper_fnc(x,scale,slope):
        return _psyfunction(x, scale, slope, 1/nafc, 0.001, sigmoid_type)

    # Fit quick function to estimate scale and slope
    param_est = curve_fit(wrapper_fnc, data[:,0], data[:,1]/data[:,2])
    
    # If initial estimates fail, try brute force method
    if param_est[0][0] != param_est[0][0]:
        warnings.warn('''Initial estimation of parameters failed. Now using brute method to try and compute initial estimates. Sit tight!''')
        
    counter = 0
    while param_est[0][0] != param_est[0][0]:
        scale_guess = (data[:,0].min() + data[:,0].max()) / 2
        slope_guess = np.linspace(-200,1000,1000)
        p0 = [scale_guess, slope_guess[counter]]
        param_est = curve_fit(wrapper_fnc, data[:,0], data[:,1]/data[:,2], p0)
        counter += 1
        
        if counter == 999:
            raise Exception('''Initial parameter estimates could not be generated!
                            While this error can arise for a number of reasons.
                            Please see API for more details.''')
    
    return param_est[0]
    

#################################################################
#  CHECK PARAMETERS ESTIMATES PROVIDED
#################################################################
def check_params(data, param_ests, nafc, batch, sigmoid_type):
    """Performs that initial parameter estimates provided 
    by user is in proper format, and assigns default values 
    where necessary.

    Keyword arguments:
    data -- m x 3 numpy array
    param_ests -- list of user specified initial parameter estimates (list)
    nafc -- N-alternative forced choice experiment (integer)
    batch -- specifies whether batch fitting data (logical)
    sigmoid_type -- sigmoid type fit to data (string)
    """
    if param_ests is not None:
        # If batch, param_ests needs to be set to None to auto generate
        # values 
        if batch is True:
            raise Warning('''Warning: User cannot provide estimates for 
                            parameters when Batch options set to TRUE!''')
        # Check whether the variable for the parameter estimates provided is of type list
        if isinstance(param_ests, list) is False:
            raise ValueError('''User Error: Please provide a argument of type list for the 
            parameter estimates of the model.''')
        # Check number of arguments provided
        if len(param_ests) < 4:
            raise ValueError('''User Error: Please provide a list of four parameter
                            estimates for the model. Set estimate to NoneType if
                            want BayesFit to generate estimate for a parameter.''')
        elif len(param_ests) > 4:
            raise ValueError('''User Error: More than 4 estimates for parameters of the model provided!''')
        # Generate value estimates for initialization to fill in those not 
        # provided
        param_initial = [0, 0, 0, 0]
        param_guess = _param_guess(data, nafc, sigmoid_type)
        scale_guess = param_guess[0]
        slope_guess = param_guess[1]
        gamma_guess = 1/nafc
        lambda_guess = 0.001
        param_tmp = [scale_guess, slope_guess, gamma_guess, lambda_guess]
        # Check whether arguments in list are of type integer or float
        NoneType = type(None)
        for i in range(0,4):
            if isinstance(param_ests[i], (int, float, NoneType)) is False:
                raise ValueError('''User Error: Please provide numerical values or NoneType 
                                 for parameter estimates of the model.''')
            # Assign default value as needed
            if param_ests[i] is None:
                param_initial[i] = param_tmp[i]
                print('Setting parameter estimates for %s to default value!')
            elif param_ests[i] is not None:
                param_initial[i] = param_ests[i]
    elif param_ests is None:
        # Get initial estimate of scale and slope
        if batch is False:
            param_guess = _param_guess(data, nafc, sigmoid_type)
            scale_guess = param_guess[0]
            slope_guess = param_guess[1]
            gamma_guess = 1/nafc
            lambda_guess = 0.001
        elif batch is True:
            # Iterate over datasets in dictionary and generate scale estimate
            # for each set
            scale_guess = dict()
            slope_guess = dict()
            gamma_guess = dict()
            lambda_guess = dict()
            for n_datasets in data:
                param_guess = _param_guess(data[n_datasets], nafc, sigmoid_type)
                scale_guess[n_datasets] = param_guess[0]
                slope_guess[n_datasets] = param_guess[1]           
                gamma_guess[n_datasets] = 1/nafc
                lambda_guess[n_datasets] = 0.001
        # Set initial parameter estimates
        param_initial = [scale_guess, slope_guess, gamma_guess, lambda_guess]

    return param_initial


#################################################################
#  CHECK WHICH PARAMETERS ARE FREE (1) VERSUS FIXED (0) 
#################################################################
def check_constraints(param_constraints):
    """Checks whether user provided list specifying which parameters 
    are to be free versus fixed is in proper format, and assigns
    default values where necessary.
    
    Keyword arguments:
    param_constraints -- specifies which parameters to estimate (list)
    """
    if param_constraints is None:
        # If no parameter constraints provided, set default values to
        # [alpha = True, beta = True, gamma = False, lambda = True]
        param_constraints = [True, True, False, False]
    elif param_constraints is not None:
        # Check whether the variable for the parameter constraints provided is of type list
        if isinstance(param_constraints, list) is False:
            raise ValueError('''User Error: Please provide a argument of type list indicating
            which parameters of the model are free (i.e., TRUE) versus fixed (i.e., FALSE).''')
        # Check whether arguments in list are of type boolean
        for i in param_constraints:
            if isinstance(i, bool) is False:
                raise ValueError('''User Error: Please provide boolean arguments for parameter '
                                constraints (i.e., free vs fixed) of the model.''')
            # Check number of arguments provided and set the remaining arguments to default values.
            if len(param_constraints) < 2:
                raise ValueError('''User Error: Please provide at least two constraints (scale; slope)
                                for parameters of the model.''')
            elif len(param_constraints) == 2:
                print('Setting parameter constraints for gamma and lambda to default values (i.e., fixed)!')
                param_constraints.append(False)
                param_constraints.append(True)
            elif len(param_constraints) == 3:
                print('Setting parameter constraints for lambda to default value (i.e., fixed)!')
                param_constraints.append(True)
            elif len(param_constraints) > 4:
                raise ValueError('''User Error: More than 4 constraints for parameters of the model provided!''')
    return param_constraints



"""
*******************************************************
*
*  BayesFit - SETUP FILE
*  
*  Version:      Version 2.4.1
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   September, 2017
*  Last updated: January 07, 2025
*
*******************************************************
"""

from setuptools import setup

setup(name='bayesfit',
      version='2.4.1',
      description='Bayesian Psychometric Curve Fitting Module',
      long_description=open('README.rst').read(),
      long_description_content_type='text/markdown',
      url='http://github.com/slugocm/BayesFit',
      author='Michael Slugocki',
      author_email='slugocm@gmail.com',
      license='Apache 2.0',
      packages=['bayesfit'],
      install_requires=['numpy', 
                        'scipy', 
                        'matplotlib'],
      zip_safe=False)


"""
*******************************************************
*
*  test_checkParams_constraints - UNIT TEST FOR TRAVIS CI
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   April 28, 2017
*  Last updated: April 29, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES 
#################################################################
import numpy as np
from .checkParams import check_constraints as _check_constraints


#################################################################
#  DEFINE FUNCTIONS USED FOR UNIT TESTING
#################################################################

def _check_constraints_none():    
    # Test cases for param_constraints is None
    param_constraints = None
    # Call function with arguments above
    _check_constraints(param_constraints)
    # Update success flag
    success = 1
    return success 
    
def _check_constraints_not_none(branch):    
    if branch == 0:
        # Test cases for param_constraints
        param_constraints = dict()
    elif branch == 1:
        # Test cases for param_constraints
        param_constraints = ['Raise', 'Exception']
    elif branch == 2:
        # Test cases for param_constraints
        param_constraints = [True]
    elif branch == 3:
        # Test cases for param_constraints
        param_constraints = [True, True]
    elif branch == 4:
        # Test cases for param_constraints
        param_constraints = [True, True, False]
    elif branch == 5:
        # Test cases for param_constraints
        param_constraints = [True, True, False, True]
    elif branch == 6:
        # Test cases for param_constraints
        param_constraints = [True, True, False, True, False]
    # Call function with arguments above
    _check_constraints(param_constraints)   
    # Update success flag
    success = 1
    return success 

#################################################################
#  UNIT TESTS
#################################################################

def test_check_params_none():
    success = _check_constraints_none()
    # Assert if exception   
    assert success == 1
    
def test_check_params_not_none_branch0():
    raised = False
    try:
        _check_constraints_not_none(0)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_params_not_none_branch1():
    raised = False
    try:
        _check_constraints_not_none(1)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_params_not_none_branch2():
    raised = False
    try:
        _check_constraints_not_none(2)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_params_not_none_branch3():
    success = _check_constraints_not_none(3)
    # Assert if exception   
    assert success == 1
    
def test_check_params_not_none_branch4():
    success = _check_constraints_not_none(4)
    # Assert if exception   
    assert success == 1
    
def test_check_params_not_none_branch5():
    success = _check_constraints_not_none(5)
    # Assert if exception   
    assert success == 1
    
def test_check_params_not_none_branch6():
    raised = False
    try:
        _check_constraints_not_none(6)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True



"""
*******************************************************
*
*  BayesFit - CORE FILE
*  
*  Version:      Version 2.4.1
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   September, 2017
*  Last updated: January 7, 2025
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES 
#################################################################
# Core modules 
import numpy as np
import re
import os 
import warnings
import matplotlib
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore", UserWarning)
    if os.environ.get('DISPLAY', '') == '':
       print('no display found. Using non-interactive Agg backend')
       matplotlib.use('Agg')
# Functions designed to check ser input
from .checkData import check_data as _check_data
from .checkLogspace import check_logspace as _check_logspace
from .checkParams import check_params as _check_params
from .checkParams import check_constraints as _check_constraints
from .checkOptions import check_options as _check_options 
from .checkPriors import check_priors as _check_priors
from .extractPriors import extract_priors as _extract_priors
# Core functions for computing likelihood/posterior
from .psyFunction import psyfunction as _psyfunction
from .genGrid import gen_grid as _gen_grid
from .gridLikelihood import grid_likelihood as _grid_likelihood
from .applyPriors import apply_priors as _apply_priors
from .calc_integral import calc_integral as _calc_integral
# Functions for extracting relevant metrics from posterior
from .extractMetrics import extract_metrics_grid as _extract_metrics_grid 
from .get_Threshold import get_threshold
# Plotting functions
from .plot_psyFcn import plot_psyfcn
from .plot_Posterior import plot_posterior
from .plot_Marginals import plot_marginals
from .plot_Priors import plot_priors


#################################################################
#  BAYESFIT CORE FUNCTION THAT ACCEPTS ARGUMENTS FROM USE 
#################################################################
def fitmodel(data,
             batch=False,
             logspace=None,
             nafc=2,
             sigmoid_type='logistic',
             param_ests=None,
             param_free=None,
             priors=None,
             threshold=0.75,
             density=100,
             ):
    """Main function used to fit psychometric functions 
    to data provided by the user.
    
    Keyword arguments:
    data -- m x 3 numpy array
    batch -- specifies whether batch fitting data (logical)
    logspace -- specifies whether to logspace data (logical)
    nafc -- N-alternative forced choice experiment (integer)
    sigmoid_type -- sigmoid type fit to data (string)
    param_ests -- estimates for each parameter (list)
    param_free -- specifies which parameters to estimate (list)
    priors -- define distributions to use as priors (list)
    threshold -- proportion correct to define threshold (float)
    density -- defines grid size across free parameters (int)
    """
    
    # Check that user provided data meets structure expected
    _check_data(data, batch)

    # Check whether x-values need to be log-spaced
    data_copy, logspace = _check_logspace(data, logspace, sigmoid_type, batch)

    # Check that parameter estimates/constraints are reasonable
    param_ests = _check_params(data_copy, param_ests, nafc, batch, sigmoid_type)

    # Check parameter constraints provided
    param_free = _check_constraints(param_free)

    # Save options used to fit functions in dictionary for reference
    options = dict()
    options['batch'] = batch
    options['logspace'] = logspace
    options['nafc'] = nafc
    options['sigmoid_type'] = sigmoid_type
    options['param_ests'] = param_ests
    options['param_free'] = param_free
    options['priors'] = priors
    options['threshold'] = threshold
    options['density'] = density

    if options['priors'] is not None:
        # Extract priors defined by user 
        options = _extract_priors(options)
        # Check priors defined by user 
        _check_priors(options)

    # Check that all options provided by user or defaults assigned are acceptable
    _check_options(options)

    # Fit models to data
    if batch is True:
        # Raise warning to inform user that only select parameters will be 
        # saved.
        warnings.warn('''\nWhen Batch mode is TRUE, Bayesfit outputs only select metrics to prevent memory issues that can arise from saving posterior and marginal distributions for a large number of datasets.''')
        tmp_metrics = dict()
        metrics = dict()
        for keys in data_copy:
            tmp_options = options.copy()
            tmp_options['param_ests'] = list()
            for i in range(0,4):
                tmp_options['param_ests'].append(options['param_ests'][i][keys])
            # Fit model
            tmp_metrics = _fit_Grid(data_copy[keys], tmp_options)
            tmp_metrics['threshold'] = get_threshold(data_copy[keys], tmp_options, tmp_metrics, tmp_options['threshold'])
            # Record only minimal values about fit to 
            # ensure memory issues are not encountered
            del tmp_metrics['posterior'], tmp_metrics['Marginals'], tmp_metrics['Marginals_X']
            metrics[keys] = tmp_metrics
    elif batch is False:
        metrics = _fit_Grid(data_copy, options)
        metrics['threshold'] = get_threshold(data_copy, options, metrics, options['threshold'])
    # Return tuple of objects
    return metrics, options


#################################################################
#  FITTING FUNCTION USING NUMERICAL INTEGRATION
#################################################################
def _fit_Grid(data, options):
    """Function used to run scripts associated with using 
    grid approximation.
    
    Keyword arguments:
    data -- m x 3 numpy array
    options -- contains all options used to fit model (dictionary)
    """    
    # Initialize metrics dictionary
    metrics = dict()
    
    # Generate grid of values for parameters (force to be equal length)
    grid = _gen_grid(data, options)

    # Save x values for grid/marginals
    metrics['Marginals_X'] = dict()
    metrics['Marginals_X']['scale'] = grid['scale']
    metrics['Marginals_X']['slope'] = grid['slope']
    metrics['Marginals_X']['gamma'] = grid['gamma']
    metrics['Marginals_X']['lambda'] = grid['lambda']
    
    # Compute likelihood
    posterior = _grid_likelihood(data, options, grid)
    metrics['likelihood_surface'] = posterior
    
    # Apply priors to likel1ihood if using Bayesian inference
    if options['priors'] is not None:
        posterior = _apply_priors(data, options, posterior, grid)
    posterior = np.exp(posterior - np.max(posterior))
    
    # Calculate integral of surface
    integral = _calc_integral(posterior, options, metrics)
    metrics['integral'] = integral
    
    # Normalize surface by integral approximated via numerical integration
    posterior /= integral
    posterior /= posterior.sum()

    # Extract relevant metrics from surface
    metrics['posterior'] = posterior
    metrics = _extract_metrics_grid(data, options, metrics, posterior, grid)
    
    # Return variables of interest 
    return metrics    



"""
*******************************************************
*
*  test_checkOptions - UNIT TEST FOR TRAVIS CI
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   April 28, 2017
*  Last updated: September 13, 2018
*
*******************************************************
"""

#################################################################
#  IMPORT MODULES 
#################################################################
import numpy as np
from .checkOptions import check_options as _check_options


#################################################################
#  DEFINE FUNCTIONS USED FOR UNIT TESTING
#################################################################

def _check_options_function(branch):    
    # Initialize options dictionary
    options = dict()
    options['batch'] = False
    options['logspace'] = False
    options['nafc'] = 2
    options['sigmoid_type'] = 'logistic'
    options['param_ests'] = [0.05, 3, 0.5, 0]
    options['param_free'] = [True, True, False, True]
    options['threshold'] = 0.75
    options['density'] = 500
    # Choose branch        
    if branch == 0:
        nothing = True
    elif branch == 1:
        options['batch'] = 'Raise Error'
    elif branch == 2:
        options['sigmoid_type'] = 'Raise Error'
    elif branch == 3:
        options['logspace'] = 'Raise Error'
    elif branch == 4:
        options['nafc'] = 12
    elif branch == 5:
        options['nafc'] = 'Raise Error'
    elif branch == 6:
        options['threshold'] = 'Raise Error'
    elif branch == 7:
        options['threshold'] = 1.5
    elif branch == 8:
        options['density'] = 'Raise Error'
    elif branch == 9:
        options['param_ests'] = dict()
    elif branch == 10:
        options['param_free'] = dict()
    # Call function with arguments above
    _check_options(options)
    # Update success flag
    success = 1
    return success 

#################################################################
#  UNIT TESTS
#################################################################

def test_check_options_branch0():
    success = _check_options_function(0)
    # Assert if exception   
    assert success == 1

def test_check_options_branch1():
    raised = False
    try:
        _check_options_function(1)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_options_branch2():
    raised = False
    try:
        _check_options_function(2)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_options_branch3():
    raised = False
    try:
        _check_options_function(3)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_options_branch4():
    raised = False
    try:
        _check_options_function(4)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_options_branch5():
    raised = False
    try:
        _check_options_function(5)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_options_branch6():
    raised = False
    try:
        _check_options_function(6)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_options_branch7():
    raised = False
    try:
        _check_options_function(7)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_options_branch8():
    raised = False
    try:
        _check_options_function(8)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_options_branch9():
    raised = False
    try:
        _check_options_function(9)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True
    
def test_check_options_branch10():
    raised = False
    try:
        _check_options_function(10)
    except:
        raised = True
    # Assert if exception flag is not raised     
    assert raised is True



"""
*******************************************************
*
*  TRAVIS CI SYNC TEST
*  
*  License:      Apache 2.0
*  Written by:   Michael Slugocki
*  Created on:   September, 2017
*  Last updated: April 28, 2018
*
*******************************************************
"""

# This file is used to test whether Travis CI is working 
# with repository. 
def func(x):
    return x + 1

def test_answer():
    assert func(4) == 5
